import urllib

import httpx
import requests
import urllib3


async def foo():
    requests.get()  # TRIO210: 4, 'requests.get'
    requests.get(...)  # TRIO210: 4, 'requests.get'
    requests.get
    print(requests.get())  # TRIO210: 10, 'requests.get'
    # fmt: off
    print(requests.get(requests.get()))# TRIO210: 10, 'requests.get' # TRIO210: 23, 'requests.get'
    # fmt: on

    requests.options()  # TRIO210: 4, 'requests.options'
    requests.head()  # TRIO210: 4, 'requests.head'
    requests.post()  # TRIO210: 4, 'requests.post'
    requests.put()  # TRIO210: 4, 'requests.put'
    requests.patch()  # TRIO210: 4, 'requests.patch'
    requests.delete()  # TRIO210: 4, 'requests.delete'
    requests.foo()

    httpx.options("")  # TRIO210: 4, 'httpx.options'
    httpx.head("")  # TRIO210: 4, 'httpx.head'
    httpx.post("")  # TRIO210: 4, 'httpx.post'
    httpx.put("")  # TRIO210: 4, 'httpx.put'
    httpx.patch("")  # TRIO210: 4, 'httpx.patch'
    httpx.delete("")  # TRIO210: 4, 'httpx.delete'
    httpx.foo()

    urllib3.request()  # TRIO210: 4, 'urllib3.request'
    urllib3.request(...)  # TRIO210: 4, 'urllib3.request'
    request()

    urllib.request.urlopen("")  # TRIO210: 4, 'urllib.request.urlopen'
    request.urlopen()  # TRIO210: 4, 'request.urlopen'
    urlopen()  # TRIO210: 4, 'urlopen'



# ARG --enable-visitor-codes-regex=(TRIO220|TRIO221)


async def foo():
    subprocess.Popen()  # TRIO220: 4, 'subprocess.Popen'
    os.system()  # TRIO221: 4, 'os.system'

    system()
    os.system.anything()
    os.anything()

    subprocess.run()  # TRIO221: 4, 'subprocess.run'
    subprocess.call()  # TRIO221: 4, 'subprocess.call'
    subprocess.check_call()  # TRIO221: 4, 'subprocess.check_call'
    subprocess.check_output()  # TRIO221: 4, 'subprocess.check_output'
    subprocess.getoutput()  # TRIO221: 4, 'subprocess.getoutput'
    subprocess.getstatusoutput()  # TRIO221: 4, 'subprocess.getstatusoutput'

    subprocess.anything()
    subprocess.foo()
    subprocess.bar.foo()
    subprocess()

    os.spawn()
    os.spawn

    os.spawnl()  # TRIO221: 4,   'os.spawnl'
    os.spawnle()  # TRIO221: 4,  'os.spawnle'
    os.spawnlp()  # TRIO221: 4,  'os.spawnlp'
    os.spawnlpe()  # TRIO221: 4, 'os.spawnlpe'
    os.spawnv()  # TRIO221: 4,   'os.spawnv'
    os.spawnve()  # TRIO221: 4,  'os.spawnve'
    os.spawnvp()  # TRIO221: 4,  'os.spawnvp'
    os.spawnvpe()  # TRIO221: 4, 'os.spawnvpe'

    # if mode is given, and is not os.P_WAIT: TRIO220
    os.spawnl(os.P_NOWAIT)  # TRIO220: 4,   'os.spawnl'
    os.spawnl(P_NOWAIT)  # TRIO220: 4,   'os.spawnl'
    os.spawnl(mode=os.P_NOWAIT)  # TRIO220: 4,   'os.spawnl'
    os.spawnl(mode=P_NOWAIT)  # TRIO220: 4,   'os.spawnl'

    # if it is P_WAIT, TRIO221
    os.spawnl(os.P_WAIT)  # TRIO221: 4,   'os.spawnl'
    os.spawnl(P_WAIT)  # TRIO221: 4,   'os.spawnl'
    os.spawnl(mode=os.P_WAIT)  # TRIO221: 4,   'os.spawnl'
    os.spawnl(mode=P_WAIT)  # TRIO221: 4,   'os.spawnl'
    # treating this as 221 to simplify code, and see no real reason not to
    os.spawnl(foo.P_WAIT)  # TRIO221: 4,   'os.spawnl'

    # other weird cases: TRIO220
    os.spawnl(0)  # TRIO220: 4,   'os.spawnl'
    os.spawnl(1)  # TRIO220: 4,   'os.spawnl'
    os.spawnl(foo())  # TRIO220: 4,   'os.spawnl'



from __future__ import annotations

import pytest


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--runfuzz", action="store_true", default=False, help="run fuzz tests"
    )
    parser.addoption(
        "--enable-visitor-codes-regex",
        default=".*",
        help="select error codes whose visitors to run.",
    )


def pytest_configure(config: pytest.Config):
    config.addinivalue_line(
        "markers", "fuzz: mark test as a slow fuzzer to not run by default"
    )


def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]):
    if config.getoption("--runfuzz"):
        # --runfuzz given in cli: do not skip fuzz tests
        return
    skip_fuzz = pytest.mark.skip(reason="need --runfuzz option to run")
    for item in items:
        if "fuzz" in item.keywords:
            item.add_marker(skip_fuzz)


@pytest.fixture()
def enable_visitor_codes_regex(request: pytest.FixtureRequest):
    return request.config.getoption("--enable-visitor-codes-regex")



import time

import trio
from trio import sleep


async def afoo():
    # only error on exactly one argument, that is 0
    await trio.sleep(0)  # error: 10
    await trio.sleep(1)
    await trio.sleep(0, 1)
    await trio.sleep(...)
    await trio.sleep()

    # don't require it being await'ed
    trio.sleep(0)  # error: 4
    trio.sleep(1)

    # don't error on other sleeps
    time.sleep(0)
    sleep(0)


# don't require being inside a function
trio.sleep(0)  # error: 0


def foo():
    # can be inside a sync function, and inside other calls
    # (though yes this is invalid code)
    trio.run(trio.sleep(0))  # error: 13



import contextlib
import contextlib as arbitrary_import_alias_for_contextlib
import functools
import os
from contextlib import asynccontextmanager
from functools import partial

import trio

# ARG --startable-in-context-manager=custom_startable_function


@asynccontextmanager
async def custom_startable_externally_tested():
    nursery.start_soon(custom_startable_function)  # error: 4
    yield


@asynccontextmanager
async def run_sampling_profiler():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(  # error: 8
            trio.run_process, ["start-sampling-profiler", str(os.getpid())]
        )
        yield

        nursery.start_soon(  # safe, it's after the first yield
            trio.run_process, ["start-sampling-profiler", str(os.getpid())]
        )


async def not_cm():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(
            trio.run_process, ["start-sampling-profiler", str(os.getpid())]
        )
        yield


class foo:
    async def __aenter__(self):
        async with trio.open_nursery() as nursery:
            nursery.start_soon(  # error: 12
                trio.run_process, ["start-sampling-profiler", str(os.getpid())]
            )
            yield


nursery = anything = custom_startable_function = anything_nursery = trio  # type: ignore


class foo2:
    _nursery = nursery

    async def __aenter__(self):
        nursery.start_soon(trio.run_process)  # error: 8
        nursery.start_soon(trio.serve_ssl_over_tcp)  # error: 8
        nursery.start_soon(trio.serve_tcp)  # error: 8
        nursery.start_soon(trio.serve_listeners)  # error: 8

        # Where does `nursery` come from in __aenter__?  Probably an attribute:
        self._nursery.start_soon(trio.run_process)  # error: 8
        self._nursery.start_soon(trio.serve_ssl_over_tcp)  # error: 8
        self._nursery.start_soon(trio.serve_tcp)  # error: 8
        self._nursery.start_soon(trio.serve_listeners)  # error: 8

        # triggers on anything whose name ends with nursery
        anything_nursery.start_soon(trio.run_process)  # error: 8
        anything.start_soon(trio.run_process)
        None.start_soon(trio.run_process)

        # explicitly check partial support
        nursery.start_soon(partial(trio.run_process))  # error: 8
        nursery.start_soon(functools.partial(trio.run_process))  # error: 8
        nursery.start_soon(anything.partial(trio.run_process))  # error: 8

        # trigger when the sensitive methods are inside a call
        nursery.start_soon(tuple(tuple(tuple(tuple(trio.run_process)))))  # error: 8
        nursery.start_soon(None, tuple(tuple(tuple(tuple(trio.run_process)))))
        nursery.start_soon(partial(print, 5))
        nursery.start_soon(partial(print, trio.run_process))  # error: 8

        serve = run_process = myfun = anything
        # error if name shared with trio
        nursery.start_soon(serve)  # error: 8
        nursery.start_soon(run_process)  # error: 8
        # don't error if a startable name is a parameter or module
        nursery.start_soon(myfun(serve=None))
        nursery.start_soon(serve.myfun)

        # doesn't support more esoteric ways of baking in the name
        nursery.start_soon(lambda x: serve(x))
        nursery.start_soon([serve])


class foo3:
    async def __aenter__(_a_parameter_not_named_self):
        nursery.start_soon(trio.run_process)  # error: 8


# might be monkeypatched onto an instance, count this as an error too
async def __aenter__():
    nursery.start_soon(trio.run_process)  # error: 4
    nursery.start_soon()  # broken code, but our analysis shouldn't crash
    nursery.cancel_scope.cancel()


@contextlib.asynccontextmanager
async def contextlib_acm():
    nursery.start_soon(trio.run_process)  # error: 4
    yield


@arbitrary_import_alias_for_contextlib.asynccontextmanager
async def contextlib_import_alias_acm():
    nursery.start_soon(trio.run_process)  # error: 4
    yield


# code coverage for non-name, non-attribute decorator
@None
async def foo4():
    ...



# specify command-line arguments to be used when testing this file.
# Test spaces in options, and trailing comma
# Cannot test newlines, since argparse splits on those if passed on the CLI
# ARG --trio200-blocking-calls=bar -> BAR, bee-> SHOULD_NOT_BE_PRINTED,bonnet ->SHOULD_NOT_BE_PRINTED,bee.bonnet->BEEBONNET,*.postwild->POSTWILD,prewild.*->PREWILD,*.*.*->TRIPLEDOT,

# don't error in sync function
def foo():
    bar()
    bee.bonnet()


async def afoo():
    bar()  # TRIO200: 4, "bar", "BAR"
    print(bar())  # TRIO200: 10, "bar", "BAR"

    # check that bee.bonnet triggers, and neither `bee` nor `bonnet`
    bee.bonnet()  # TRIO200: 4, "bee.bonnet", "BEEBONNET"

    # check wildcard support
    bar.postwild()  # TRIO200: 4, "*.postwild", "POSTWILD"
    prewild.anything()  # TRIO200: 4, "prewild.*", "PREWILD"
    a.b.c()  # TRIO200: 4, "*.*.*", "TRIPLEDOT"

    # don't error when it's not a call
    bar
    bee.bonnet
    boo(bar=None)

    # it won't catch esoteric ways of calling the function
    [bee.bonnet][0]()
    x = bee.bonnet
    x()

    # don't error inside lambda
    lambda: bar()

    # check that states are properly set/reset on nested functions
    def bar():
        bar()

        async def bar():
            bar()  # TRIO200: 12, "bar", "BAR"

    bar()  # TRIO200: 4, "bar", "BAR"

    # don't error on directly awaited expressions
    # https://github.com/Zac-HD/flake8-trio/issues/85
    await bar()
    print(await bar())

    # error on not directly awaited expressions
    await print(bar())  # TRIO200: 16, "bar", "BAR"
    await (foo() if bar() else foo())  # TRIO200: 20, "bar", "BAR"
    await bee.bonnet(bar())  # TRIO200: 21, "bar", "BAR"

    # known false alarms
    await (x := bar())  # TRIO200: 16, "bar", "BAR"
    await (bar() if True else foo())  # TRIO200: 11, "bar", "BAR"
    await (await bee.bonnet() or bar())  # TRIO200: 33, "bar", "BAR"
    await ((lambda: bee.bonnet()) and bar())  # TRIO200: 38, "bar", "BAR"

    # check that errors are enabled again
    bar()  # TRIO200: 4, "bar", "BAR"



import trio

# ARG --startable-in-context-manager=foo


async def foo(task_status):
    ...


async def bar(task_status):  # error: 0, "bar"
    async def barbar(task_status):  # error: 4, "barbar"
        ...

    def sync_bar():
        async def barbar(task_status):  # error: 8, "barbar"
            ...


async def foo2(task_status=trio.TASK_STATUS_IGNORED):  # error: 0, "foo2"
    ...


async def foo3(*, task_status):  # error: 0, "foo3"
    ...


async def foo4(task_status, /):  # error: 0, "foo4"
    ...


async def foo5(*task_status):
    ...


async def foo6(**task_status):
    ...


def sync():
    async def sync_async(task_status):  # error: 4, "sync_async"
        ...



from __future__ import annotations

import ast
import copy
import itertools
import os
import re
import site
import subprocess  # noqa: S404
import sys
import tokenize
import unittest
from collections import deque
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any, DefaultDict

import pytest
from flake8 import __version_info__ as flake8_version_info
from flake8.options.manager import OptionManager

# import trio  # type: ignore
from hypothesis import HealthCheck, given, settings
from hypothesmith import from_grammar, from_node

from flake8_trio import ERROR_CLASSES, Error, Plugin, Statement

# TODO: Move test_eval files into a separate directory
trio_test_files_regex = re.compile(r"trio\d.*.py")

test_files: list[tuple[str, str]] = sorted(
    (os.path.splitext(f)[0].upper(), f)
    for f in os.listdir("tests")
    if re.match(trio_test_files_regex, f)
)


class ParseError(Exception):
    ...


# flake8 6 added a required named parameter formatter_names
def _default_option_manager():
    kwargs = {}
    if flake8_version_info[0] >= 6:
        kwargs["formatter_names"] = ["default"]
    return OptionManager(version="", plugin_versions="", parents=[], **kwargs)


# check for presence of _pyXX, skip if version is later, and prune parameter
def check_version(test: str):
    python_version = re.search(r"(?<=_PY)\d*", test)
    if python_version:
        version_str = python_version.group()
        major, minor = version_str[0], version_str[1:]
        v_i = sys.version_info
        if (v_i.major, v_i.minor) < (int(major), int(minor)):
            pytest.skip(f"python version {v_i} smaller than {major}, {minor}")


ERROR_CODES = {
    err_code: err_class
    for err_class in ERROR_CLASSES
    for err_code in err_class.error_codes.keys()
}


@pytest.mark.parametrize(("test", "path"), test_files)
def test_eval(test: str, path: str):
    # version check
    check_version(test)
    test = test.split("_")[0]

    parsed_args = []

    # only enable the tested visitor to save performance and ease debugging
    # if a test requires enabling multiple visitors they specify a
    # `# ARG --enable-vis...` that comes later in the arg list, overriding this
    if test in ERROR_CODES:
        parsed_args = [f"--enable-visitor-codes-regex={test}"]

    expected: list[Error] = []

    with open(os.path.join("tests", path), encoding="utf-8") as file:
        lines = file.readlines()

    for lineno, line in enumerate(lines, start=1):
        # interpret '\n' in comments as actual newlines
        line = line.replace("\\n", "\n")

        line = line.strip()

        # add command-line args if specified with #ARG
        if reg_match := re.search(r"(?<=ARG ).*", line):
            parsed_args.append(reg_match.group().strip())

        # skip commented out lines
        if not line or line[0] == "#":
            continue

        # get text between `error:` and (end of line or another comment)
        k = re.findall(r"(error|TRIO...):([^#]*)(?=#|$)", line)

        for err_code, err_args in k:
            try:
                # Append a bunch of empty strings so string formatting gives garbage
                # instead of throwing an exception
                try:
                    args = eval(  # noqa: S307
                        f"[{err_args}]",
                        {
                            "lineno": lineno,
                            "line": lineno,
                            "Statement": Statement,
                            "Stmt": Statement,
                        },
                    )
                except NameError:
                    print(f"failed to eval on line {lineno}", file=sys.stderr)
                    raise

            except Exception as e:
                print(f"lineno: {lineno}, line: {line}", file=sys.stderr)
                raise e

            if args:
                col, *args = args
            else:
                col = 0
            assert isinstance(
                col, int
            ), f'invalid column "{col}" @L{lineno}, in "{line}"'

            if err_code == "error":
                err_code = test
            error_class = ERROR_CODES[err_code]
            message = error_class.error_codes[err_code]
            try:
                expected.append(Error(err_code, lineno, int(col), message, *args))
            except AttributeError as e:
                msg = (
                    f"Line {lineno}: Failed to format\n {message!r}\n" f'"with\n{args}'
                )
                raise ParseError(msg) from e

    assert parsed_args, "no parsed_args"
    assert expected, f"failed to parse any errors in file {path}"

    plugin = read_file(path)
    assert_expected_errors(plugin, *expected, args=parsed_args)


# Codes that are supposed to also raise errors when run on sync code, and should
# be excluded from the SyncTransformer check.
# Expand this list when adding a new check if it does not care about whether the code
# is asynchronous or not.
error_codes_ignored_when_checking_transformed_sync_code = {
    "TRIO100",
    "TRIO101",
    "TRIO103",
    "TRIO104",
    "TRIO105",
    "TRIO106",
    "TRIO111",
    "TRIO112",
    "TRIO115",
    "TRIO116",
}


class SyncTransformer(ast.NodeTransformer):
    def visit_Await(self, node: ast.Await):
        return self.generic_visit(node.value)

    def replace_async(self, node: ast.AST, target: type[ast.AST]) -> ast.AST:
        node = self.generic_visit(node)
        newnode = target()
        newnode.__dict__ = node.__dict__
        return newnode

    def visit_AsyncFunctionDef(self, node: ast.AST):
        return self.replace_async(node, ast.FunctionDef)

    def visit_AsyncWith(self, node: ast.AST):
        return self.replace_async(node, ast.With)

    def visit_AsyncFor(self, node: ast.AST):
        return self.replace_async(node, ast.For)


@pytest.mark.parametrize(("test", "path"), test_files)
def test_noerror_on_sync_code(test: str, path: str):
    if any(e in test for e in error_codes_ignored_when_checking_transformed_sync_code):
        return
    with tokenize.open(f"tests/{path}") as f:
        source = f.read()
    tree = SyncTransformer().visit(ast.parse(source))

    ignored_codes_regex = (
        "(?!("
        + "|".join(error_codes_ignored_when_checking_transformed_sync_code)
        + "))"
    )
    assert_expected_errors(
        Plugin(tree),
        args=[f"--enable-visitor-codes-regex={ignored_codes_regex}"],
    )


def read_file(test_file: str):
    filename = Path(__file__).absolute().parent / test_file
    return Plugin.from_filename(str(filename))


def initialize_options(plugin: Plugin, args: list[str] | None = None):
    om = _default_option_manager()
    plugin.add_options(om)
    plugin.parse_options(om.parse_args(args=(args if args else [])))


def assert_expected_errors(
    plugin: Plugin,
    *expected: Error,
    args: list[str] | None = None,
):
    # initialize default option values
    initialize_options(plugin, args)

    errors = sorted(e for e in plugin.run())
    expected_ = sorted(expected)

    print_first_diff(errors, expected_)
    assert_correct_lines_and_codes(errors, expected_)
    assert_correct_columns(errors, expected_)
    assert_correct_args(errors, expected_)

    # full check
    assert errors == expected_

    # test tuple conversion and iter types
    assert_tuple_and_types(errors, expected_)


def print_first_diff(errors: Sequence[Error], expected: Sequence[Error]):
    first_error_line: list[Error] = []
    first_expected_line: list[Error] = []
    for err, exp in zip(errors, expected):
        if err == exp:
            continue
        if not first_error_line or err.line == first_error_line[0]:
            first_error_line.append(err)
        if not first_expected_line or exp.line == first_expected_line[0]:
            first_expected_line.append(exp)

    if first_expected_line != first_error_line:
        print(
            "First lines with different errors",
            f"  actual: {[e.cmp() for e in first_error_line]}",
            f"expected: {[e.cmp() for e in first_expected_line]}",
            "",
            sep="\n",
            file=sys.stderr,
        )


def assert_correct_lines_and_codes(errors: Iterable[Error], expected: Iterable[Error]):
    MyDict = DefaultDict[int, DefaultDict[str, int]]
    # Check that errors are on correct lines
    all_lines = sorted({e.line for e in (*errors, *expected)})
    error_dict: MyDict = DefaultDict(lambda: DefaultDict(int))
    expected_dict = copy.deepcopy(error_dict)

    for e in errors:
        error_dict[e.line][e.code] += 1
    for e in expected:
        expected_dict[e.line][e.code] += 1

    any_error = False
    for line in all_lines:
        if error_dict[line] == expected_dict[line]:
            continue
        for code in sorted({*error_dict[line], *expected_dict[line]}):
            if not any_error:
                print(
                    "Lines with different # of errors:",
                    "-" * 38,
                    f"| line | {'code':7} | actual | expected |",
                    sep="\n",
                    file=sys.stderr,
                )
                any_error = True

            print(
                f"| {line:4}",
                f"{code}",
                f"{error_dict[line][code]:6}",
                f"{expected_dict[line][code]:8} |",
                sep=" | ",
                file=sys.stderr,
            )
    assert not any_error


def assert_correct_columns(errors: Iterable[Error], expected: Iterable[Error]):
    # check errors have correct columns
    col_error = False
    for err, exp in zip(errors, expected):
        assert err.line == exp.line
        if err.col != exp.col:
            if not col_error:
                print("Errors with same line but different columns:", file=sys.stderr)
                print("| line | actual | expected |", file=sys.stderr)
                col_error = True
            print(
                f"| {err.line:4} | {err.col:6} | {exp.col:8} |",
                file=sys.stderr,
            )
    assert not col_error


def assert_correct_args(errors: Iterable[Error], expected: Iterable[Error]):
    # check errors have correct messages
    args_error = False
    for err, exp in zip(errors, expected):
        assert (err.line, err.col, err.code) == (exp.line, exp.col, exp.code)
        if err.args != exp.args:
            if not args_error:
                print(
                    "Errors with different args:",
                    "-" * 20,
                    sep="\n",
                    file=sys.stderr,
                )
                args_error = True
            print(
                f"*    line: {err.line:3} differs\n",
                f"  actual: {err.args}\n",
                f"expected: {exp.args}\n",
                "-" * 20,
                file=sys.stderr,
            )
    assert not args_error


def assert_tuple_and_types(errors: Iterable[Error], expected: Iterable[Error]):
    def info_tuple(error: Error):
        try:
            return tuple(error)
        except IndexError:
            print(
                "Failed to format error message",
                f"line: {error.line}",
                f"col: {error.col}",
                f"code: {error.code}",
                f"args: {error.args}",
                f'format string: "{ERROR_CODES[error.code].error_codes[error.code]}"',
                sep="\n    ",
                file=sys.stderr,
            )
            raise

    for err, exp in zip(errors, expected):
        err_msg = info_tuple(err)
        for err, type_ in zip(err_msg, (int, int, str, type)):
            assert isinstance(err, type_)
        assert err_msg == info_tuple(exp)


def test_107_permutations():
    """
    since each test is so fast, and there's so many permutations, manually doing
    the permutations in a single test is much faster than the permutations from using
    pytest parametrization - and does not clutter up the output massively.

    generates code that looks like this, where a block content of `None` means the
    block is excluded:

    async def foo():
        try:
            await foo() | ...
        except ValueError:
            await foo() | ... | raise | return | None
        except SyntaxError:
            await foo() | ... | raise | return | None
        except:
            await foo() | ... | raise | return | None
        else:
            await foo() | ... | return | None
        finally:
            await foo() | ... | return | None
    """
    plugin = Plugin(ast.AST())
    initialize_options(plugin, args=["--enable-visitor-codes-regex=TRIO107"])

    check = "await foo()"

    # loop over all the possible content of the different blocks
    for try_, exc1, exc2, bare_exc, else_, finally_ in itertools.product(
        (check, "..."),  # try_
        (check, "...", "raise", "return", None),  # exc1
        (check, "...", "raise", "return", None),  # exc2
        (check, "...", "raise", "return", None),  # bare_exc
        (check, "...", "return", None),  # else_
        (check, "...", "return", None),  # finally_
    ):
        # exclude duplicate tests where there's a second exception block but no first
        if exc1 is None and exc2 is not None:
            continue

        # syntax error if there's no exception block but there's finally and/or else
        if exc1 is exc2 is bare_exc is None and (finally_ is None or else_ is not None):
            continue

        function_str = f"async def foo():\n  try:\n    {try_}\n"

        for arg, val in {
            "except ValueError": exc1,
            "except SyntaxError": exc2,
            "except": bare_exc,
            "else": else_,
            "finally": finally_,
        }.items():
            if val is not None:
                function_str += f"  {arg}:\n    {val}\n"

        tree = ast.parse(function_str)

        # not a type error per se, but it's pyright warning about assigning to a
        # protected class member - hence we silence it with a `type: ignore`.
        plugin._tree = tree  # type: ignore
        errors = list(plugin.run())

        if (
            # return in exception
            "return" in (exc1, exc2, bare_exc)
            # exception and finally doesn't checkpoint, checkpoint in try might not run
            or ("..." in (exc1, exc2, bare_exc) and finally_ != check)
            # no checkpoints in normal control flow
            or check not in (try_, finally_, else_)
            # return in else|finally w/o checkpoint before
            or ("return" in (else_, finally_) and check not in (else_, try_))
            # return in finally with no bare exception, checkpoint in try might not run
            or (finally_ == "return" and bare_exc is None)
        ):
            assert errors, "# missing alarm:\n" + function_str
        else:
            assert not errors, "# false alarm:\n" + function_str


def test_114_raises_on_invalid_parameter(capsys: pytest.CaptureFixture[str]):
    plugin = Plugin(ast.AST())
    # flake8 will reraise ArgumentError as SystemExit
    for arg in "blah.foo", "foo*", "*":
        with pytest.raises(SystemExit):
            initialize_options(plugin, args=[f"--startable-in-context-manager={arg}"])
        out, err = capsys.readouterr()
        assert not out
        assert f"{arg!r} is not a valid method identifier" in err


def test_200_options(capsys: pytest.CaptureFixture[str]):
    plugin = Plugin(ast.AST())
    for i, arg in (0, "foo"), (2, "foo->->bar"), (2, "foo->bar->fee"):
        with pytest.raises(SystemExit):
            initialize_options(plugin, args=[f"--trio200-blocking-calls={arg}"])
        out, err = capsys.readouterr()
        assert not out, out
        assert all(word in err for word in (str(i), arg, "->"))


def _test_trio200_from_config_common(tmp_path: Path) -> str:
    tmp_path.joinpath(".flake8").write_text(
        """
[flake8]
trio200-blocking-calls =
  other -> async,
  sync_fns.* -> the_async_equivalent,
select = TRIO200
"""
    )
    tmp_path.joinpath("example.py").write_text(
        """
import sync_fns

async def foo():
    sync_fns.takes_a_long_time()
"""
    )
    return (
        "./example.py:5:5: TRIO200 User-configured blocking sync call sync_fns.* "
        "in async function, consider replacing with the_async_equivalent.\n"
    )


def test_200_from_config_flake8_internals(
    tmp_path: Path, capsys: pytest.CaptureFixture[str]
):
    # abuse flake8 internals to avoid having to use subprocess
    # which breaks breakpoints and hinders debugging
    # TODO: fixture (?) to change working directory

    err_msg = _test_trio200_from_config_common(tmp_path)
    # replace ./ with tmp_path/
    err_msg = str(tmp_path) + err_msg[1:]

    from flake8.main.cli import main

    main(
        argv=[
            str(tmp_path / "example.py"),
            "--append-config",
            str(tmp_path / ".flake8"),
        ]
    )
    out, err = capsys.readouterr()
    assert not err
    assert err_msg == out


def test_200_from_config_subprocess(tmp_path: Path):
    err_msg = _test_trio200_from_config_common(tmp_path)
    res = subprocess.run(  # noqa: S603,S607
        ["flake8"], cwd=tmp_path, capture_output=True
    )
    assert not res.stderr
    assert res.stdout == err_msg.encode("ascii")


# from https://docs.python.org/3/library/itertools.html#itertools-recipes
def consume(iterator: Iterable[Any]):
    deque(iterator, maxlen=0)


@pytest.mark.fuzz()
class TestFuzz(unittest.TestCase):
    @settings(max_examples=1_000, suppress_health_check=[HealthCheck.too_slow])
    @given((from_grammar() | from_node()).map(ast.parse))
    def test_does_not_crash_on_any_valid_code(self, syntax_tree: ast.AST):
        # TODO: figure out how to get unittest to play along with pytest options
        # so `--enable-visitor-codes-regex` can be passed through.
        # Though I barely notice a difference manually changing this value, or even
        # not running the plugin at all, so overhead looks to be vast majority of runtime
        enable_visitor_codes_regex = ".*"

        # Given any syntatically-valid source code, the checker should
        # not crash.  This tests doesn't check that we do the *right* thing,
        # just that we don't crash on valid-if-poorly-styled code!
        plugin = Plugin(syntax_tree)
        initialize_options(
            plugin, [f"--enable-visitor-codes-regex={enable_visitor_codes_regex}"]
        )

        consume(plugin.run())


def _iter_python_files():
    # Because the generator isn't perfect, we'll also test on all the code
    # we can easily find in our current Python environment - this includes
    # the standard library, and all installed packages.
    for base in sorted(set(site.PREFIXES)):
        for dirname, _, files in os.walk(base):
            for f in files:
                if f.endswith(".py"):
                    yield Path(dirname) / f


@pytest.mark.fuzz()
def test_does_not_crash_on_site_code(enable_visitor_codes_regex: str):
    for path in _iter_python_files():
        try:
            plugin = Plugin.from_filename(str(path))
            initialize_options(
                plugin, [f"--enable-visitor-codes-regex={enable_visitor_codes_regex}"]
            )
            consume(plugin.run())
        except Exception as err:
            raise AssertionError(f"Failed on {path}") from err



from urllib3 import PoolManager


async def foo():
    foo = PoolManager()
    foo.request("OPTIONS")  # TRIO211: 4, 'foo.request'
    foo.request("GET")  # TRIO211: 4, 'foo.request'
    foo.request("HEAD")  # TRIO211: 4, 'foo.request'
    foo.request("POST")  # TRIO211: 4, 'foo.request'
    foo.request("PUT")  # TRIO211: 4, 'foo.request'
    foo.request("DELETE")  # TRIO211: 4, 'foo.request'
    foo.request("TRACE")  # TRIO211: 4, 'foo.request'
    foo.request("CONNECT")  # TRIO211: 4, 'foo.request'
    foo.request("PATCH")  # TRIO211: 4, 'foo.request'

    foo.request("PUT", ...)  # TRIO211: 4, 'foo.request'
    request("PUT")
    foo.request()
    foo.request(..., "PUT")
    foo.request("put")  # TRIO211: 4, 'foo.request'
    foo.request("sPUTnik")
    foo.request(None)
    foo.request(put)
    foo.request(PUT)
    foo.request("REQUEST")  # not an HTTP verb



from __future__ import annotations

import ast

from flake8.main.application import Application

from flake8_trio import Statement, Visitor107_108, fnmatch_qualified_name


def dec_list(*decorators: str) -> ast.Module:
    source = ""
    for dec in decorators:
        source += f"@{dec}\n"
    source += "async def f():\n  bar()"
    return ast.parse(source)


def wrap(decorators: tuple[str, ...], decs2: str) -> str | None:
    tree = dec_list(*decorators)
    assert isinstance(tree.body[0], ast.AsyncFunctionDef)
    return fnmatch_qualified_name(tree.body[0].decorator_list, decs2)


def test_basic():
    assert wrap(("foo",), "foo")
    assert wrap(("foo", "bar"), "foo")
    assert wrap(("bar", "foo"), "foo")

    assert not wrap(("foo",), "foob")


def test_dotted():
    assert wrap(("foo.bar",), "foo.bar")

    assert not wrap(("foo.bar",), "foo")
    assert not wrap(("foo.bar",), "bar")

    assert not wrap(("foo",), "foo.bar")
    assert not wrap(("bar",), "foo.bar")

    assert not wrap(("foo.bar.jane",), "foo.bar")
    assert not wrap(("foo.bar",), "foo.bar.jane")
    assert not wrap(("jane.foo.bar",), "foo.bar")
    assert not wrap(("foo.bar",), "jane.foo.bar")


def test_multidotted():
    assert wrap(("foo.bar.jane",), "foo.bar.jane")
    assert not wrap(("foo.bar",), "foo.bar.jane")
    assert not wrap(("foo.bar.jane",), "foo.bar")


def test_wildcard():
    assert wrap(("foo",), "*")
    assert wrap(("foo.bar",), "*")
    assert not wrap(("foo",), "foo.*")
    assert wrap(("foo.bar",), "foo.*")
    assert not wrap(("bar.foo",), "foo.*")

    assert wrap(("foo",), "foo*")
    assert wrap(("foobar",), "foo*")
    assert wrap(("foobar.bar",), "foo*")
    assert wrap(("foo.bar",), "*.bar")


def test_at():
    assert wrap(("foo",), "@foo")
    assert wrap(("foo.bar",), "@foo.bar")


def test_calls():
    assert wrap(("foo()",), "@foo")
    assert wrap(("foo(1, 2, *x, **y)",), "@foo")
    assert wrap(("foo.bar()",), "@foo.bar")
    assert wrap(("foo.bar(1, 2, *x, **y)",), "@foo.bar")


def test_pep614():
    # Just don't crash and we'll be good.
    assert not wrap(("(any, expression, we, like)",), "no match here")


common_flags = ["--select=TRIO", "tests/trio_options.py"]


def test_command_line_1(capfd):
    Application().run(common_flags + ["--no-checkpoint-warning-decorators=app.route"])
    assert capfd.readouterr() == ("", "")


expected_out = (
    "tests/trio_options.py:2:1: TRIO107 "
    + Visitor107_108.error_codes["TRIO107"].format(
        "exit", Statement("function definition", 2)
    )
    + "\n"
)


def test_command_line_2(capfd):
    Application().run(common_flags + ["--no-checkpoint-warning-decorators=app"])
    assert capfd.readouterr() == (expected_out, "")


def test_command_line_3(capfd):
    Application().run(common_flags)
    assert capfd.readouterr() == (expected_out, "")



import math
from math import inf

import trio


async def foo():
    # These examples are probably not meant to ever wake up:
    await trio.sleep(100000)  # error: 10

    # and these ones _definitely_ never wake up:
    await trio.sleep(float("inf"))  # error: 10
    await trio.sleep(math.inf)  # error: 10
    await trio.sleep(inf)  # error: 10
    # 'inf literal' overflow trick
    await trio.sleep(1e999)  # error: 10

    await trio.sleep(86399)
    await trio.sleep(86400)
    await trio.sleep(86400.01)  # error: 10
    await trio.sleep(86401)  # error: 10

    await trio.sleep(-1)  # will raise a runtime error from Trio
    await trio.sleep(0)  # handled by different check

    # don't evaluate expressions
    await trio.sleep(10**10)
    await trio.sleep(86400 + 1)
    await trio.sleep(foo())
    await trio.sleep(86400 + foo())  # type: ignore
    await trio.sleep(86400 + ...)  # type: ignore
    await trio.sleep("hello")
    await trio.sleep(...)


# does not require the call to be awaited, nor in an async fun
trio.sleep(86401)  # error: 0
# also checks that we don't break visit_Call
trio.run(trio.sleep(86401))  # error: 9
